/*---------------------------------------------------------------
 * ESP32_WS2812B_TCPServer.ino
 * AUTHOR - baelzabubba
 * DATE 05/08/2022
 * GNU Creative Commons License
 * baelzabubbaAThotmailDOTcom
 * --------------------------------------------------------------
 * If you use this script or a derivitive of it, please cite your
 * source.
 * 
 * This script sets up a TCPIP Server on the ESP32 to recieve and
 * parse commands from a TCPIP Client that are used to set the RGB
 * Colour and State of the individual LED on a WS2812B strip.
 ----------------------------------------------------------------*/
 
#include <FastLED.h>
#include "FS.h"
#include <SPI.h>
#include <SD.h>
#include <ArduinoJson.h>
#include <WiFi.h>

#define LED_PIN       32
#define NUM_LEDS      125
#define BRIGHTNESS    125
#define LED_TYPE      WS2812B
#define COLOR_ORDER   GRB

CRGB leds[NUM_LEDS];

#define UPDATES_PER_SECOND  100

/* data structures */
struct Config{
  char ssid[32];
  char password[29];
  char ep[32];
  char port[32];
};

struct State{
  int RGBValue[3];
  byte StateValues[25];
};

Config config;      //WiFi Configuration
const char *cfg = "/etc/config.json";
bool sdLoad = false;

State state[5];
 
WiFiServer wifiServer(80);

CRGBPalette16 currentPalette;
TBlendType    currentBlending;

extern CRGBPalette16 myRedWhiteBluePalette;
extern const TProgmemPalette16 myRedWhiteBluePalette_p PROGMEM;

void setup() {
  delay(3000);  //power-up safety delay
  
  Serial.begin(115200);

  FastLED.addLeds<LED_TYPE, LED_PIN, COLOR_ORDER>(leds, NUM_LEDS).setCorrection( TypicalLEDStrip );
  FastLED.setBrightness(  BRIGHTNESS );

  currentPalette = RainbowColors_p;
  currentBlending = LINEARBLEND;    
  
  SD.begin();
  if(!SD.begin()){
    Serial.println("Card Mount Failure");
    sdLoad = false;
    return;
  }

  Serial.println("Initializing SD card ... ");
  if(!SD.begin()){
    Serial.println("ERROR - SD card initialization failed!");
    sdLoad = false;
    return;
  }
  
  File file = SD.open(cfg);
  if(!file){
    Serial.println("File doesn't exist");
  } else {
    Serial.println("File already exists");    
    sdLoad = true;
  }
  file.close();

  if(sdLoad)
  {
      Serial.println(F("Loading configuration..."));
      loadConfiguration(cfg, config);
  }
  delay(1000);
  WiFi.begin(config.ssid, config.password);
  Serial.print("Connecting to WiFi");
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.print(".");
  }
  Serial.println();
  Serial.println("Connected to the WiFi network");
  Serial.println(WiFi.localIP());
  Serial.print("Port: ");
  Serial.println(config.port);
  //printStruct();
 
  wifiServer.begin(atoi(config.port));
  Serial.println("Ready to connect.");
  showData();
}
 
void loop() {
  WiFiClient client = wifiServer.available();
  uint size=0;
  String nString;
  if (client) {
    while (client.connected()) {
      while ((size = client.available())>0) {
        char c = client.read();
        if(c != '\n'){
          nString += (String)c;
        } else {
          //echoSerial(nString);
          parseSerial(nString);
          //Serial.println(nString);
          nString = "";
          client.write("ok\n");
        }       
      }
      delay(10);      
    }
    client.stop();    
  }
}

void loadConfiguration(const char *filename, Config &config) {
  File cfgFile = SD.open(filename);

  StaticJsonDocument<192> doc;

  DeserializationError error = deserializeJson(doc, cfgFile);

  if (error) {
    Serial.print("deserializeJson() failed: ");
    Serial.println(error.c_str());
    return;
  }

  strlcpy(config.ssid, doc["SSID"] | "********", sizeof(config.ssid));
  strlcpy(config.password, doc["pass"] | "********", sizeof(config.password));
  strlcpy(config.ep, doc["ep"] | "192.168.1.100", sizeof(config.ep));
  strlcpy(config.port, doc["port"] | "4096", sizeof(config.port));
  cfgFile.close();
  return;
}

/* function to output Config struct to file */
void saveConfiguration(const char *filename, Config &config) {
  File cfgFile = SD.open(filename, FILE_WRITE);
  
  StaticJsonDocument<128> doc;

  doc["SSID"] = config.ssid;
  doc["pass"] = config.password;
  doc["ep"] = config.ep;
  doc["port"] = config.port;

  serializeJson(doc, cfgFile);
  cfgFile.close();
  return;
}

void printStruct() {
  Serial.print("ssid: ");
  Serial.println(config.ssid);
  Serial.print("password: ");
  Serial.println(config.password);
  Serial.print("Endpoint: ");
  Serial.println(config.ep);
  Serial.print("Port: ");
  Serial.println(config.port);
}

void echoSerial(String msg) {
  Serial.println(msg);
}

void parseSerial(String msg) {
  int L = msg.indexOf("=");
  String command = msg.substring(0, L);
  String values = msg.substring(L+1);
  if(command == "C")
  {    
    // get the colour data from the message
    L = values.indexOf(",");
    
    int index = values.substring(0,L).toInt();
    int parent = index;
    for(int i = 0; i < 3; i++)
    {
      int next = values.indexOf(",",L+1);
      index = values.substring(L+1, next).toInt();
      L = next;
      state[parent].RGBValue[i] = index;
    }
  }
  else if(command == "S") //get the state data from the message
  {
    L = values.indexOf(",");
    int index = values.substring(0,L).toInt();
    int parent = index;
    for(int i = 0; i < 25; i++)
    {
      int next = values.indexOf(",",L+1);
      index = values.substring(L+1, next).toInt();
      L = next;
      state[parent].StateValues[i] = index;
    }
  } else if ( command = "D") { showData();}
}

void showData()
{
  for(int p = 0; p < 5; p++)
  {
    Serial.print("Player Index: "); Serial.println(p);
    Serial.print("Colours = ");
    for(int c = 0; c < 3; c++)
    {
      Serial.print(state[p].RGBValue[c]);
      if(c < 2) {Serial.print(",");} else {Serial.println();}
    }
    Serial.print("State = ");
    for(int s = 0; s < 25; s++)
    {
      Serial.print(state[p].StateValues[s]);
      if(s < 24) {Serial.print(",");} else {Serial.println();}      
    }
  }

  /*
   * Once we have shown the data on Serial, it is time to send the data out to the 
   * WS2812B LED light strip.
   */
   FastLED.delay(1000 / UPDATES_PER_SECOND);
   FastLED.clear();
   for(int p = 0; p < 5; p++)
   {
      int L = 25;
      int TC = p * L;
      for(int s = 0; s < L; s++)      
      {
          int C = s + TC;
          if(state[p].StateValues[s] == 1)
          {
            Serial.print("LED "); Serial.println(C);
            leds[C] = CRGB(state[p].RGBValue[0], state[p].RGBValue[1], state[p].RGBValue[2]);
          }
      }
   }   
   FastLED.show();
}
